#!/usr/bin/env python

import sys
import six
import nkf
import yaml
import opt
import term
import key
import rect
import txbuf
import txview
#import dbg
#import subprocess as sp

frm_info = {
	'': '   ',
	'a': '+-|',
	'b': ' -|',
	'c': '==#',
}

class Empty:
	pass

def get_tx_rect(o):
	return (o.x, o.y, o.w, o.h)

def get_frm_rect(o):
	r = get_tx_rect(o)

	(bx, by) = o.frm.bdr	

	ha = o.frm.ha
	(x, y, w, h) = r
	if ha == 'l':
		r = (x, y, w+bx, h)
	elif ha == 'r':
		r = (x-bx, y, w+bx, h)
	else: # c
		r = (x-int(bx/2), y, w+bx, h)	

	va = o.frm.va
	(x, y, w, h) = r
	if va == 'u':
		r = (x, y, w, h+by)
	elif va == 'd':
		r = (x, y-by, w, h+by)
	else: # c
		r = (x, y-int(by/2), w, h+by)

	(x, y, w, h) = r
	r = (x-1, y-1, w+2, h+2)
	return r

def p_on_frm(p, o):
	rects = rect.frame_rects( get_frm_rect(o) )
	for r in rects:
		if rect.in_rect(p, r):
			return True
	return False

def align_str(s, w, a):
	m = w - nkf.str_width(s)
	if a == 'l':
		s += ' ' * m
	elif a == 'r':
		s = ' ' * m + s
	else: # c
		m2 = int(m/2)
		s = ' ' * m2 + s + ' ' * (m-m2)
	return s

def draw_frm_buf_atts(o, buf, atts):
	cs = frm_info.get( o.frm.kind )
	(x, y, w, h) = get_frm_rect(o)

	s = cs[0] + cs[1]*(w-2) + cs[0]
	buf.set(x, y, s, atts)
	buf.set(x, y+h-1, s, atts)
	for i in range(1,h-1):
		buf.set(x, y+i, cs[2], atts)
		buf.set(x+w-1, y+i, cs[2], atts)

def draw_o_buf_atts(o, buf, atts, do_frm, frm_atts):
	for i in range(o.h):
		s = align_str( o.s[i], o.w, o.a )
		buf.set( o.x, o.y+i, s, atts )
	if do_frm:
		draw_frm_buf_atts(o, buf, frm_atts)

def new(lst, nkf_opt):
	e = Empty()

	rects = list( map( get_frm_rect, lst ) )
	r = rect.rects_contain_rect(rects)
	view = txview.new(r, nkf_opt)

	(x, y, w, h) = view.get_rect()
	e.cur = (x, y)

	def frm_visible(o):
		if o.frm.kind != '':
			return True
		if o != e.sel:
			return False
		return e.mode in ('sel_2', 'sel_frm')

	def get_o_rect(o):
		return get_frm_rect(o) if frm_visible(o) else get_tx_rect(o)

	def get_atts(o, buf):
		if o == e.sel:
			if e.mode in ('sel', 'move', 'frm_align'):
				return [ buf.rev ]
			elif e.mode in ('sel_align'):
				return [ buf.uline ]
		return []
		
	def get_frm_atts(o, buf):
		if o == e.sel:
			if e.mode in ('sel', 'move', 'sel_3', 'resize'):
				return [ buf.rev ]
			if e.mode in ('sel_2', 'sel_frm'):
				return [ buf.uline ]
		return []

	def draw_o_buf(o, buf):
		atts = get_atts(o, buf)
		do_frm = frm_visible(o)
		frm_atts = get_frm_atts(o, buf)
		draw_o_buf_atts(o, buf, atts, do_frm, frm_atts)

	def draw_r(r):
		buf = txbuf.new(r)

		for o in lst:
			ro = get_o_rect(o)
			if rect.is_area( rect.op_and(ro, r) ):
				draw_o_buf(o, buf)

		# cursor
		if e.mode == 'normal':
			if rect.in_rect(e.cur, r):
				(x, y) = e.cur
				buf.set(x, y, '+')
		buf.adjust()

		(rx, ry, rw, rh) = r
		for i in range(rh):
			y = ry + i
			(x, s) = buf.get_line(y)
			view.out(x, y, s)

	def draw_r2(bak, r):
		sc = view.scroll(r)

		vr = view.get_rect()
		rects = [ vr ]

		if not sc:
			bak = rect.op_and(bak, vr)
			is_bak = rect.is_area(bak)

			r = rect.op_and(r, vr)
			is_r = rect.is_area(r)

			if is_bak and is_r:
				rects = rect.gather(bak, r)
			elif is_bak:
				rects = [ bak ]
			elif is_r:
				rects = [ r ]
			else:
				return
		for r in rects:
			draw_r(r)

	def draw_o(o):
		draw_r( get_o_rect(o) )

	def move_cur(dx, dy):
		bak = rect.p_wh_to( e.cur, (1, 1) )
		e.cur = rect.p_add( e.cur, (dx, dy) )
		r = rect.p_wh_to( e.cur, (1, 1) )
		draw_r2(bak, r)

	def get_o_under_cursor():
		for o in lst:
			if rect.in_rect( e.cur, get_o_rect(o) ):
				return o
		return None

	def set_mode(mode):
		if mode == 'sel':
			e.sel = get_o_under_cursor()
		o = e.sel
		bak = get_o_rect(o) if o else None

		e.mode = mode
		if mode == 'normal':
			e.sel = None

		if o and bak:
			draw_r2( bak, get_o_rect(o) )

	def normal(k):
		d = key.allow(k)
		(dx, dy) = key.dxy( k, [0, 0] )

		(x, y, w, h) = view.get_rect()
		(cx, cy) = e.cur
		if d == 'pu':
			dy = y - cy
		elif d == 'pd':
			dy = y+h-1 - cy
		elif d == 'pl':
			dx = x - cx
		elif d == 'pr':
			dx = x+w-1 - cx

		if dx != 0 or dy != 0:
			move_cur(dx, dy)

	def resize(k):
		o = e.sel
		bak = get_o_rect(o)
		(bx, by) = o.frm.bdr
		d = key.allow(k)
		if d == 'u' and by > 0:
			by -= 1
		elif d == 'd':
			by += 1
		elif d == 'l' and bx > 0:
			bx -= 1
		elif d == 'r':
			bx += 1
		if (bx, by) != o.frm.bdr:
			o.frm.bdr = [bx, by] # list
			draw_r2( bak, get_o_rect(o) )

	def sel_align(k):
		o = e.sel
		lst = ['l', 'c', 'r']
		o.a = lst[ ( lst.index(o.a) + 1 ) % len(lst) ]
		draw_o(o)

	def move(k):
		o = e.sel
		(dx, dy) = key.dxy( k, [0, 0] )
		bak = get_o_rect(o)
		(o.x, o.y) = (o.x + dx, o.y + dy)
		draw_r2( bak, get_o_rect(o) )
		move_cur(dx, dy) # hidden

	def sel_frm(k):
		o = e.sel
		lst = sorted( frm_info.keys() )
		o.frm.kind = lst[ ( lst.index(o.frm.kind) + 1 ) % len(lst) ]
		draw_o(o)

	def frm_align(k):
		o = e.sel
		bak = get_o_rect(o)
		d = key.allow(k)
		if d in 'ud':
			lst = ['u', 'c', 'd']
			o.frm.va = lst[ ( lst.index(o.frm.va) + 1 ) % len(lst) ]
			draw_r2( bak, get_o_rect(o) )
		elif d in 'lr':
			lst = ['l', 'c', 'r']
			o.frm.ha = lst[ ( lst.index(o.frm.ha) + 1 ) % len(lst) ]
			draw_r2( bak, get_o_rect(o) )

	def has_bdr(k):
		(bx, by) = e.sel.frm.bdr
		return bx > 0 or by > 0

	hdl = {
		'normal': {
			'ok': {
				'f': ( lambda k: get_o_under_cursor() != None ),
				True: 'sel' },
			'allow': normal },
		'sel': { 'ok': 'sel_2', 'allow': [ 'move', move ] },
		'move': { 'ok': 'normal', 'allow': move },
		'sel_2': {
			'ok': {
				'f': ( lambda k: p_on_frm(e.cur, e.sel) ),
				True: 'sel_3',
				False: {
					'f': ( lambda k: len(e.sel.s) > 1 ),
					True: 'sel_align',
					False: 'normal' }
				},
			'allow': [ 'sel_frm', sel_frm ] },
		'sel_frm': { 'ok': 'normal', 'allow': sel_frm },
		'sel_align': { 'ok': 'normal', 'allow': sel_align },
		'sel_3': {
			'ok': { 'f': has_bdr, True: 'frm_align', False: 'normal' },
			'allow': [ 'resize', resize ] },
		'resize': { 'ok': 'normal', 'allow': resize },
		'frm_align': { 'ok': 'normal', 'allow': frm_align },
	}

	def call(o, k):
		if callable(o):
			return o(k)
		if type(o) == str:
			return set_mode(o)
		if type(o) == list:
			for e in o:
				ret = call(e, k)
			return ret
		if type(o) == dict:
			if 'f' in o:
				ret = call( o.get('f'), k )
				return call( o.get(ret), k )
			if key.is_ok(k):
				return call( o.get('ok'), k )
			if key.allow(k) in 'udlr':
				return call( o.get('allow'), k )
		return None

	def get_key(klst):
		key.init()
		while True:
			k = key.get()
			if k in klst:
				break
			o = hdl.get(e.mode)
			call(o, k)
		key.fini()
	e.get_key = get_key

	e.fini = lambda : view.fini()

	e.mode = 'normal'
	e.sel = None

	draw_r(r)

	return e

def to_empty(d):
	e = Empty()
	for (k, v) in d.items():
		setattr(e, k, v)
	return e

def cvt_data_load(o, y):
	if isinstance(o, six.string_types):
		o = o.strip().split('\n')
	if type(o) == list:
		o = { 's': o }
	# dict
	if 'xy' not in o:
		o['xy'] = [ 1, y+1 ]
		y += len( o.get('s') ) + 1
	if 'a' not in o:
		o['a'] = 'l'
	if 'frm' not in o:
		o['frm'] = { 'kind': '', 'va': 'c', 'ha': 'c', 'bdr': [ 0, 0 ] }

	e = to_empty(o)
	if type(e.s) != list:
		e.s = [ e.s ]
	(e.x, e.y) = e.xy
	e.w = max( map( nkf.str_width, e.s ) )
	e.h = len(e.s)
	e.frm = to_empty(e.frm)

	return (e, y)

def cvt_data_save(e):
	if len(e.s) == 1:
		e.s = e.s[0]
	d = {}
	d['s'] = e.s
	d['xy'] = [ e.x, e.y ]
	d['a'] = e.a

	frm = e.frm
	f = {}
	f['kind'] = frm.kind
	f['va'] = frm.va
	f['ha'] = frm.ha
	f['bdr'] = frm.bdr

	d['frm'] = f
	return d

def txt_out(lst, nkf_opt):
	rects = list( map( get_frm_rect, lst ) )
	r = rect.rects_contain_rect(rects)
	buf = txbuf.new(r)	
	for o in lst:
		draw_o_buf_atts(o, buf, [], o.frm.kind, [])

	s = nkf.enc( buf.txt_out() )
	if nkf_opt != '-u':
		s = nkf.cvt(s, nkf_opt)
	nkf.put_stdout(s + '\n')

if __name__ == "__main__":
	(s, nkf_opt) = nkf.to_str( nkf.get_stdin() )
	lst = yaml.load(s)

	y = 0
	(bak, lst) = (lst, [])
	for o in bak:
		(o, y) = cvt_data_load(o, y)
		lst.append(o)

	o = opt.get('-o', '')
	if not o:
		o = nkf_opt

	if opt.index('-txt') >= 0:
		txt_out(lst, o)
		sys.exit(0)

	# edit
	a = new(lst, o)
	a.get_key( ['q'] )
	a.fini()

	lst = list( map( cvt_data_save, lst ) )
	u8 = yaml.dump( lst, default_flow_style=False, allow_unicode=True, encoding='utf-8' )
	nkf.put_stdout( nkf.cvt(u8, nkf_opt) if nkf_opt != '-u' else u8 )
# EOF
